# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import asyncio
import typing as ty
from contextlib import asynccontextmanager
from datetime import timedelta
from types import TracebackType

from ai4s.jobq.entities import Response, Task


class Envelope(ty.Protocol):
    """Represents a message received from the queue.

    .. note:: **Backend-specific behavior of ``replace()``**

       - **Storage Queue**: atomically updates the message content in-place
         (uses ``update_message``).
       - **Service Bus**: no-op—Service Bus has no in-place update API.
         The message will be re-delivered with its original content after
         the lock expires.  Users who need retry budgets on Service Bus
         should track retries in their own application code.
    """

    @property
    def task(self) -> Task: ...

    @property
    def id(self) -> str: ...

    @property
    def lock_lost_event(self) -> asyncio.Event: ...

    async def delete(self, success: bool, error: str | None = None) -> None: ...

    async def requeue(self) -> None: ...

    async def replace(self, task: Task) -> None: ...

    async def reply(self, response: Response) -> None: ...

    async def cancel_heartbeat(self) -> None: ...


class JobQBackendWorker(ty.Protocol):
    def receive_message(
        self, visibility_timeout: timedelta, with_heartbeat: bool = False, **kwargs
    ) -> ty.AsyncContextManager[Envelope]: ...

    async def push(self, task: Task) -> str: ...


class JobQBackend(ty.Protocol):
    @property
    def name(self) -> str: ...

    async def __aenter__(self) -> "JobQBackend": ...

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None: ...

    async def push(self, task: Task) -> str: ...

    def receive_message(
        self, visibility_timeout: timedelta, with_heartbeat: bool = False, **kwargs
    ) -> "ty.AsyncContextManager[Envelope]": ...

    async def create(self, exist_ok: bool = True) -> None: ...

    async def clear(self) -> None: ...

    async def __len__(self) -> int: ...

    def generate_sas(self, ttl: timedelta) -> str: ...

    async def peek(self, n: int = 1, as_json: bool = False) -> ty.Any: ...

    async def get_result(self, session_id: str, timeout: timedelta | None = None) -> Response: ...

    @asynccontextmanager
    async def get_worker_interface(self, **kwargs) -> ty.AsyncGenerator[JobQBackendWorker, None]:
        yield self



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""
This is an entrypoint to the orchestration features of ai4s.jobq.

If users import from this file, the autocomplete will be uncluttered and auto-documenting.
"""

from ai4s.jobq.orchestration.manager import batch_enqueue, get_results, launch_workers
from ai4s.jobq.work import WorkSpecification

__all__ = ["WorkSpecification", "batch_enqueue", "get_results", "launch_workers"]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import os

import asyncclick as click

SKILL_NAME = "ai4s-jobq-cli"

# Extra usage examples for commands where the Click help text alone
# doesn't demonstrate key options.
COMMAND_EXAMPLES = {
    "ai4s-jobq push": [
        "ai4s-jobq $JOBQ push -c 'echo hello'  # push a single command",
        "cat tasks.txt | ai4s-jobq $JOBQ push   # push commands from stdin",
    ],
    "ai4s-jobq worker": [
        "ai4s-jobq $JOBQ worker -n 4 --time-limit 12h  # 4 parallel workers, 12h limit",
    ],
    "ai4s-jobq peek": [
        "ai4s-jobq $JOBQ peek -n 5 --json  # peek at 5 messages as JSON",
    ],
}


def _format_param(param):
    """Format a single Click parameter (option or argument) as a string."""
    if isinstance(param, click.Option):
        decls = ", ".join(f"`{d}`" for d in param.opts + param.secondary_opts)
        parts = [f"  - {decls}"]
        if not param.is_flag:
            type_name = getattr(param.type, "name", None)
            if type_name:
                type_name = type_name.upper()
                if type_name not in ("TEXT", "STRING", "<LAMBDA>"):
                    parts.append(f" `{type_name}`")
        if param.help:
            parts.append(f": {param.help}")
        if param.default is not None and param.default != () and not param.is_flag:
            parts.append(f" (default: `{param.default}`)")
        if param.required:
            parts.append(" **(required)**")
        return "".join(parts)
    if isinstance(param, click.Argument):
        name = param.human_readable_name
        parts = [f"  - `{name}`"]
        if param.required:
            parts.append(" **(required)**")
        if param.nargs == -1:
            parts.append(" (multiple)")
        return "".join(parts)
    return None


def _is_group(cmd):
    """Check if a command is a group (works with asyncclick patched classes)."""
    return hasattr(cmd, "list_commands") and hasattr(cmd, "get_command")


def _walk_commands(group, prefix="ai4s-jobq"):
    """Yield (full_name, command) tuples by walking the Click command tree."""
    if not _is_group(group):
        return
    ctx = click.Context(group, info_name=prefix)
    for name in sorted(group.list_commands(ctx)):
        cmd = group.get_command(ctx, name)
        if cmd is None or getattr(cmd, "hidden", False):
            continue
        # Skip the copilot-skill group itself from the skill reference
        if name == "copilot-skill":
            continue
        full_name = f"{prefix} {name}"
        yield full_name, cmd
        if _is_group(cmd):
            yield from _walk_commands(cmd, prefix=full_name)


def _generate_skill_md(cli_group):
    """Generate SKILL.md content from the Click command tree."""
    lines = []

    # YAML frontmatter
    lines.append("---")
    lines.append("name: ai4s-jobq-cli")
    lines.append("description: >-")
    lines.append(
        "  Authoritative reference for the ai4s-jobq CLI for managing"
        " distributed job queues on Azure Storage Queues and Azure Service Bus."
    )
    lines.append("---")
    lines.append("")
    lines.append("  When constructing ai4s-jobq commands, you MUST look up the exact option names")
    lines.append("  from this document. Do NOT guess or infer option names from other CLIs.")
    lines.append("  Every option for every command is listed below.")
    lines.append("")

    # Body
    lines.append("# ai4s-jobq CLI Reference")
    lines.append("")
    lines.append("This skill provides a complete reference of all `ai4s-jobq` CLI commands,")
    lines.append("their options, and arguments. Use it to help users construct correct")
    lines.append("`ai4s-jobq` command invocations.")
    lines.append("")
    lines.append(
        "`ai4s-jobq` is a distributed job queue built on Azure Storage Queues and"
        " Azure Service Bus. Users push tasks; workers pull and process them asynchronously."
    )
    lines.append("")
    lines.append("## Queue specification")
    lines.append("")
    lines.append(
        "All commands require a `SERVICE/QUEUE_NAME` argument (or the `JOBQ_STORAGE_QUEUE`"
        " environment variable):"
    )
    lines.append("")
    lines.append("```")
    lines.append("# Azure Storage Queue")
    lines.append("ai4s-jobq mystorageaccount/my-queue <command>")
    lines.append("")
    lines.append("# Azure Service Bus")
    lines.append("ai4s-jobq sb://myservicebus/my-queue <command>")
    lines.append("```")
    lines.append("")
    lines.append("## Tips for agents")
    lines.append("")
    lines.append(
        "**Only use options and arguments documented below.** Do not invent or guess"
        " option names. Every available option for each command is listed explicitly."
    )
    lines.append("")

    for full_name, cmd in _walk_commands(cli_group):
        lines.append(f"## `{full_name}`")
        lines.append("")
        help_text = (cmd.help or cmd.get_short_help_str(limit=300) or "").strip()
        if help_text:
            lines.append(help_text)
            lines.append("")

        # Show subcommands summary for groups
        if _is_group(cmd):
            ctx = click.Context(cmd, info_name=full_name)
            sub_names = cmd.list_commands(ctx)
            visible = [
                n for n in sub_names if not getattr(cmd.get_command(ctx, n), "hidden", False)
            ]
            if visible:
                lines.append("**Subcommands:** " + ", ".join(f"`{n}`" for n in sorted(visible)))
                lines.append("")

        # Options and arguments (filter out help)
        params = [p for p in cmd.params if p.name != "help"]
        formatted_params = [_format_param(p) for p in params]
        formatted_params = [f for f in formatted_params if f]
        if formatted_params:
            lines.append("**Parameters:**")
            lines.append("")
            lines.extend(formatted_params)
            lines.append("")

        # Usage examples
        examples = COMMAND_EXAMPLES.get(full_name)
        if examples:
            lines.append("**Examples:**")
            lines.append("")
            lines.append("```")
            lines.extend(examples)
            lines.append("```")
            lines.append("")

    return "\n".join(lines)


SKILL_DIR = os.path.join(os.path.expanduser("~"), ".copilot", "skills", SKILL_NAME)
DISABLED_STAMP = os.path.join(SKILL_DIR, ".disabled")


@click.group("copilot-skill")
def skill_file_cmd():
    """Manage the GitHub Copilot skill for the ai4s-jobq CLI."""


@skill_file_cmd.command("install")
def skill_file_install():
    """Install the ai4s-jobq Copilot skill to ~/.copilot/skills/."""
    import logging
    import shutil
    from importlib.resources import files as pkg_files

    from ai4s.jobq import __version__

    LOG = logging.getLogger("ai4s.jobq")

    # Remove disabled marker if present
    if os.path.exists(DISABLED_STAMP):
        os.remove(DISABLED_STAMP)

    # Try to use bundled skill data from package
    bundled = pkg_files("ai4s.jobq.data.skill").joinpath(SKILL_NAME)
    if bundled.is_dir() and bundled.joinpath("SKILL.md").is_file():
        shutil.rmtree(SKILL_DIR, ignore_errors=True)
        shutil.copytree(str(bundled), SKILL_DIR)
        # Write version stamp
        with open(os.path.join(SKILL_DIR, ".jobq-version"), "w") as f:
            f.write(__version__)
        LOG.info(f"Skill installed to {SKILL_DIR}")
    else:
        # Fallback: generate from live Click tree (no docs references)
        LOG.warning(
            "Bundled skill data not found. Generating from live CLI tree. "
            "Run 'python scripts/build_skill.py' to include documentation references."
        )
        from ai4s.jobq.cli import main

        content = _generate_skill_md(main)
        os.makedirs(SKILL_DIR, exist_ok=True)
        with open(os.path.join(SKILL_DIR, "SKILL.md"), "w") as f:
            f.write(content)
        with open(os.path.join(SKILL_DIR, ".jobq-version"), "w") as f:
            f.write(__version__)
        LOG.info(f"Skill file written to {os.path.join(SKILL_DIR, 'SKILL.md')}")

    click.echo(f"Copilot skill installed to {SKILL_DIR}")


@skill_file_cmd.command("clear")
def skill_file_clear():
    """Remove the ai4s-jobq Copilot skill and prevent auto-reinstall."""
    import shutil

    if os.path.exists(SKILL_DIR):
        shutil.rmtree(SKILL_DIR)
    # Leave a disabled marker so auto-install doesn't re-create it
    os.makedirs(SKILL_DIR, exist_ok=True)
    with open(DISABLED_STAMP, "w") as f:
        f.write(
            "Skill disabled by 'ai4s-jobq copilot-skill clear'."
            " Run 'ai4s-jobq copilot-skill install' to re-enable.\n"
        )
    click.echo("Skill removed. Run 'ai4s-jobq copilot-skill install' to re-enable.")


@skill_file_cmd.command("list")
def skill_file_list():
    """List all files that would be installed by the Copilot skill."""
    from importlib.resources import files as pkg_files

    bundled = pkg_files("ai4s.jobq.data.skill").joinpath(SKILL_NAME)
    if not bundled.is_dir():
        click.echo("No bundled skill data found.", err=True)
        raise SystemExit(1)

    def _walk(traversable, prefix=""):
        for item in sorted(traversable.iterdir(), key=lambda x: x.name):
            path = os.path.join(prefix, item.name) if prefix else item.name
            if item.is_file():
                click.echo(path)
            elif item.is_dir():
                _walk(item, path)

    _walk(bundled)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import errno
import logging
import os
import signal
from glob import glob

LOG = logging.getLogger(__name__)


def send_signal(pid: int, signal_to_send: signal.Signals):
    try:
        os.kill(pid, signal_to_send)
        LOG.info("%s signal sent to process %s.", signal_to_send.name, pid)
    except ProcessLookupError:
        LOG.info("No process with PID %s found.", pid)
    except PermissionError:
        LOG.info("Permission denied to send %s to process %s.", signal_to_send.name, pid)


def send_signal_by_glob(pattern: str, signal_to_send: signal.Signals):
    for pid_file_path in glob(pattern):
        pid = PidFile.check_pid(pid_file=pid_file_path)
        if pid is not None:
            send_signal(pid=pid, signal_to_send=signal_to_send)


class PidFile:
    """
    A context manager that creates pid files in given path, that are acting as a lock mechanism.
    Adapted from https://stackoverflow.com/a/23837022
    """

    def __init__(self, path: str):
        self.pid_file = path
        self.pidfd: int | None = None

    def __enter__(self):
        try:
            parent_dir = os.path.dirname(self.pid_file)
            os.makedirs(parent_dir, exist_ok=True)
            self.pidfd = os.open(self.pid_file, os.O_CREAT | os.O_WRONLY | os.O_EXCL)
            LOG.info("locked pidfile %s", self.pid_file)
        except OSError as e:
            if e.errno == errno.EEXIST:  # File exists
                pid = self.check_pid(pid_file=self.pid_file)
                if pid is not None:  # This pid file is already in use
                    self.pidfd = None
                    raise ProcessRunningException(
                        "process already running in %s as pid %s", self.pid_file, pid
                    ) from e
                os.remove(self.pid_file)
                LOG.warning("removed stale lockfile %s", self.pid_file)
                self.pidfd = os.open(self.pid_file, os.O_CREAT | os.O_WRONLY | os.O_EXCL)
            else:
                raise

        os.write(self.pidfd, str(os.getpid()).encode())
        os.close(self.pidfd)
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback) -> bool:
        if exception_type is None:
            self._remove()
            return True
        if issubclass(exception_type, ProcessRunningException):
            return False
        if self.pidfd:
            self._remove()
        return False

    def _remove(self):
        LOG.info("removed pidfile %s", self.pid_file)
        os.remove(self.pid_file)

    @staticmethod
    def check_pid(pid_file: str) -> int | None:
        with open(pid_file) as f:
            try:
                pid_str = f.read()
                pid = int(pid_str)
            except ValueError:
                LOG.info("not an integer: %s", pid_str)
                return None
            try:
                # Killing a process with signal `0` does nothing to a running process.
                # It raises an OSError exception with errno=ESRCH if the process does not exist.
                # https://stackoverflow.com/a/13402639
                os.kill(pid, 0)
            except OSError:
                LOG.info("can't deliver signal to %s", pid)
                return None
            else:
                return pid


class ProcessRunningException(BaseException):
    """
    There is an attempt to create a pid file;
    however this pid file is already in use by another process.
    """



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""
Events like VM preemptions are announced by Azure through the scheduled events API.
This module provides a mechanism to poll the scheduled events API and yield the events as they arrive.

https://learn.microsoft.com/en-us/azure/virtual-machines/windows/scheduled-events

Example usage:

```python
import ai4s.jobq.scheduled_events

shutdown_event = asyncio.Event()
async with PreemptionEventHandler(event):
    # wait until preempted
    await shutdown_event.wait()
```
"""

import asyncio
import contextlib
import logging
import os
from dataclasses import dataclass

import aiohttp

from ai4s.jobq import JobQ

LOG = logging.getLogger(__name__)


@dataclass
class ScheduledEvent:
    EventId: str
    EventStatus: str
    EventType: str
    ResourceType: str
    Resources: list[str]
    NotBefore: str
    Description: str
    EventSource: str
    DurationInSeconds: int


class ScheduledEventHandler:
    def __init__(self, callback, poll_interval_seconds: int = 1):
        self._polling_task: asyncio.Task | None = None
        self._poll_interval_seconds = poll_interval_seconds
        self.__callback = callback

    async def __aenter__(self):
        if not running_on_azure():
            return self

        self._polling_task = asyncio.create_task(self._poll_scheduled_events())
        self._session = aiohttp.ClientSession()
        await self._session.__aenter__()

        LOG.info("Started polling for scheduled events.")
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        if self._polling_task:
            await self._session.__aexit__(exc_type, exc_value, traceback)
            self._polling_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._polling_task
            LOG.info("Stopped polling for scheduled events.")

    async def _poll_scheduled_events(self):
        """Continuously polls the scheduled events API and logs events."""
        try:
            while True:
                events = await self._get_scheduled_events()
                await self.__callback(events)
                await asyncio.sleep(self._poll_interval_seconds)
        except asyncio.CancelledError:
            LOG.info("Polling for scheduled events has been cancelled.")

    async def _get_scheduled_events(self):
        """Fetches scheduled events from the Azure metadata service."""
        # Call 169.254.169.254, a special IP address that Azure VMs can use to access the metadata service.
        # This IP address is local to the VM and is not routable.
        # Docs: https://learn.microsoft.com/en-us/azure/virtual-machines/windows/scheduled-events
        try:
            async with self._session.get(
                "http://169.254.169.254/metadata/scheduledevents",
                headers={"Metadata": "true"},
                params={"api-version": "2020-07-01"},
            ) as resp:
                data = await resp.json()
        except Exception as e:
            LOG.warning(f"Failed to fetch scheduled events: {e}")
            return []
        return [ScheduledEvent(**event) for event in data["Events"]]


class PreemptionEventHandler(ScheduledEventHandler):
    """A simple handler to manage shutdown events."""

    def __init__(
        self,
        shutdown_event: asyncio.Event,
        queue: JobQ | None = None,
        environment_name: str | None = None,
        *args,
        **kwargs,
    ):
        self.shutdown_event = shutdown_event
        self.queue = queue
        self.environment_name = environment_name

        super().__init__(self.__callback, *args, **kwargs)

        self._scheduled_event_initiated_by_us = False

        truish = ["1", "true", "yes", "t", "y"]
        self.disabled = os.environ.get("JOBQ_DISABLE_SCHEDULED_EVENTS", "0").lower() in truish

    async def __callback(self, events: list[ScheduledEvent]):
        is_preempted = False
        for event in events:
            if event.EventType == "Preempt":
                LOG.warning(
                    f"Preemption event detected! Event ID: {event.EventId!r}, "
                    f"Not Before: {event.NotBefore!r}, Description: {event.Description!r}",
                    extra={
                        "event": "preemption_detected",
                    },
                )
                is_preempted = True

        if is_preempted:
            if not self.disabled:
                self.shutdown_event.set()
                self._scheduled_event_initiated_by_us = True
        elif self.shutdown_event.is_set() and self._scheduled_event_initiated_by_us:
            LOG.warning(
                "Preemption event is not present anymore, apparently it's obsolete.",
                extra={
                    "event": "preemption_cleared",
                },
            )
            if not self.disabled:
                self.shutdown_event.clear()
                self._scheduled_event_initiated_by_us = False


def running_on_azure() -> bool:
    # To check if we are on Azure, we check if an environment variable from
    # Azure Batch (backend of AzureML) or from
    # native Azure Batch have AZ_BATCH_TASK_ID is set
    return "AZ_BATCHAI_CLUSTER_NAME" in os.environ or "AZ_BATCH_TASK_ID" in os.environ



import asyncio
from datetime import timedelta

import pytest

from ai4s.jobq.entities import EmptyQueue, LockLostError
from ai4s.jobq.jobq import JobQ, _is_async_callable


@pytest.mark.asyncio
async def test_async(mocker, async_queue, azurite_connstr):
    """
    check whether we can queue a command and execute it in a worker
    """
    commands = []

    async def callback(cmd: str):
        commands.append(cmd)

    await async_queue.push("echo hello world")

    async with JobQ.from_connection_string("jobs", connection_string=azurite_connstr) as q_worker:
        success0 = await q_worker.pull_and_execute(callback)
        with pytest.raises(EmptyQueue):
            await q_worker.pull_and_execute(callback)

    assert success0
    assert commands == ["echo hello world"]


@pytest.mark.asyncio
async def test_heartbeat_lock_lost(async_queue, azurite_connstr):
    """When the heartbeat detects that the lock is lost, the task should be cancelled
    and the worker should return False without trying to settle the message."""
    callback_started = asyncio.Event()
    callback_cancelled = asyncio.Event()

    async def callback(cmd: str):
        callback_started.set()
        try:
            # Long sleep that should be cancelled when the lock is lost.
            await asyncio.sleep(3600)
        except asyncio.CancelledError:
            callback_cancelled.set()
            raise

    await async_queue.push("echo hello world")

    async with JobQ.from_connection_string("jobs", connection_string=azurite_connstr) as q_worker:
        # Access the underlying storage queue client so we can sabotage the message.
        backend = q_worker._client
        assert backend.queue_client is not None

        original_update = backend.queue_client.update_message

        call_count = 0

        async def sabotaged_update(message, *args, **kwargs):
            nonlocal call_count
            call_count += 1
            if call_count >= 2:
                # Corrupt the pop receipt so the real Azure API rejects it.
                kwargs["pop_receipt"] = "invalid-pop-receipt"
            return await original_update(message, *args, **kwargs)

        backend.queue_client.update_message = sabotaged_update

        with pytest.raises(LockLostError):
            await q_worker.pull_and_execute(
                callback, visibility_timeout=timedelta(seconds=2), with_heartbeat=True
            )

    assert callback_cancelled.is_set(), "Callback should have been cancelled"

    # The message should NOT have been deleted or deadlettered — it must still
    # be in the queue so another worker can pick it up.  Wait for the
    # visibility timeout to expire so it becomes visible again.
    await asyncio.sleep(3)
    async with JobQ.from_connection_string("jobs", connection_string=azurite_connstr) as q2:
        msgs = await q2.peek(1)
        assert len(msgs) >= 1, "Message should still be in the queue"
    """
    check whether we can execute a long-running command by sending continuous heartbeat updates
    """

    job_duration = 10

    async def callback(cmd: str):
        await asyncio.sleep(job_duration)

    await async_queue.push("echo hello world")

    async with JobQ.from_connection_string("jobs", connection_string=azurite_connstr) as q_worker:
        # set minimum visibility timeout, which should cause expiry
        success = await q_worker.pull_and_execute(
            callback, visibility_timeout=timedelta(seconds=2), with_heartbeat=True
        )
    assert success


@pytest.mark.asyncio
async def test_is_aio_callable():
    async def async_fn():
        pass

    def sync_fn():
        pass

    class A:
        async def __call__(self):
            pass

    class B:
        def __call__(self):
            pass

    assert _is_async_callable(async_fn)
    assert not _is_async_callable(sync_fn)
    assert _is_async_callable(A())
    assert not _is_async_callable(B())



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import os
from functools import lru_cache
from logging import getLogger

from azure.core.credentials import TokenCredential
from azure.core.credentials_async import AsyncTokenCredential

LOG = getLogger(__name__)


@lru_cache(2)
def get_token_credential() -> AsyncTokenCredential:
    """
    Make a good guess which token credential to use.

    We avoid using DefaultAzureCredential, since on sandboxes, this picks up the managed identity.
    We check whether DEFAULT_IDENTITY_CLIENT_ID is set, in which case this is
    likely a user-assigned managed identity, eg of an aml cluster.

    Finally, we return the AzureCliCredential.
    """
    from azure.identity.aio import (
        AzureCliCredential,
        ManagedIdentityCredential,
    )

    if "DEFAULT_IDENTITY_CLIENT_ID" in os.environ:
        LOG.info("Authenticating with ManagedIdentityCredential()")
        return ManagedIdentityCredential(client_id=os.environ["DEFAULT_IDENTITY_CLIENT_ID"])
    LOG.info("Authenticating with AzureCliCredential()")
    return AzureCliCredential()


@lru_cache(2)
def get_sync_token_credential() -> TokenCredential:
    """
    Make a good guess which token credential to use.

    We avoid using DefaultAzureCredential, since on sandboxes, this picks up the managed identity.
    We check whether DEFAULT_IDENTITY_CLIENT_ID is set, in which case this is
    likely a user-assigned managed identity, eg of an aml cluster.

    Finally, we return the AzureCliCredential.
    """
    from azure.identity import (
        AzureCliCredential,
        ManagedIdentityCredential,
    )

    if "DEFAULT_IDENTITY_CLIENT_ID" in os.environ:
        LOG.info("Authenticating with ManagedIdentityCredential()")
        return ManagedIdentityCredential(client_id=os.environ["DEFAULT_IDENTITY_CLIENT_ID"])
    LOG.info("Authenticating with AzureCliCredential()")
    return AzureCliCredential()



"""Sphinx extension that replaces HTML-producing directives with plain alternatives.

Used when building markdown docs (SPHINX_MARKDOWN_BUILDER=1) so that
directives like ``.. prompt::`` produce clean markdown instead of raw HTML.
"""

from typing import ClassVar

from docutils import nodes  # type: ignore[import-untyped]
from docutils.parsers.rst import Directive  # type: ignore[import-untyped]


class MarkdownPrompt(Directive):
    """Replace ``.. prompt::`` with a plain code-block."""

    optional_arguments = 3  # e.g. bash $ auto
    has_content = True
    option_spec: ClassVar[dict] = {"substitutions": lambda x: x}

    def run(self):
        code = "\n".join(self.content)
        node = nodes.literal_block(code, code)
        node["language"] = "bash"
        return [node]


def setup(app):
    app.add_directive("prompt", MarkdownPrompt, override=True)
    return {"version": "0.1", "parallel_read_safe": True}



import contextlib
import os
from pathlib import Path

import pytest
from azure.storage.blob import BlobProperties
from azure.storage.blob.aio import BlobServiceClient

from ai4s.jobq.blob import BlobContainer


@pytest.mark.asyncio
async def test_blob_processor(mocker, tmp_path) -> None:
    container = "mock0container"
    blob_port = os.environ.get("BLOB_PORT", "10000")
    bsc = BlobServiceClient.from_connection_string(
        f"DefaultEndpointsProtocol=http;AccountName=devstoreaccount1;AccountKey=Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw==;BlobEndpoint=http://127.0.0.1:{blob_port}/devstoreaccount1;"
    )
    with contextlib.suppress(Exception):
        await bsc.create_container(container)
    mocker.patch("ai4s.jobq.blob.BlobServiceClient", return_value=bsc)
    async with BlobContainer(
        storage_account="mock0storage0account",
        container=container,
        credential="mock0credential",
    ) as bp:
        local_path = tmp_path.joinpath("mock0blob")
        local_path.write_text("mock0content")
        await bp.upload_file(local_path, "remote_path")
        assert await bsc.get_blob_client(container, "remote_path").exists()
        assert await bp.blob_exists("remote_path")
        downloaded_path = await bp.download_file("remote_path", tmp_path)
        assert downloaded_path.startswith(str(tmp_path))
        assert Path(downloaded_path).read_text() == "mock0content"

        downloaded_path2, properties = await bp.download_file_with_properties(
            "remote_path", tmp_path
        )
        assert Path(downloaded_path2).read_text() == Path(downloaded_path).read_text()
        assert isinstance(properties, BlobProperties)
        # check a random property
        assert properties.last_modified is not None

        subfolder = tmp_path.joinpath("subfolder")
        subfolder.mkdir(parents=True, exist_ok=True)
        local_path1 = subfolder.joinpath("mock1file")
        local_path2 = subfolder.joinpath("mock2file")
        local_path1.write_text("mock1content")
        local_path2.write_text("mock2content")
        await bp.upload_from_folder(subfolder, "remote_path")
        assert await bsc.get_blob_client(container, "remote_path/mock1file").exists()
        assert await bsc.get_blob_client(container, "remote_path/mock2file").exists()


@pytest.mark.asyncio
async def test_blob_stash(mocker) -> None:
    container = "mock0container"
    blob_port = os.environ.get("BLOB_PORT", "10000")
    bsc = BlobServiceClient.from_connection_string(
        f"DefaultEndpointsProtocol=http;AccountName=devstoreaccount1;AccountKey=Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw==;BlobEndpoint=http://127.0.0.1:{blob_port}/devstoreaccount1;"
    )
    with contextlib.suppress(Exception):
        await bsc.create_container(container)
    mocker.patch("ai4s.jobq.blob.BlobServiceClient", return_value=bsc)
    async with BlobContainer(
        storage_account="mock0storage0account",
        container=container,
        credential="mock0credential",
    ) as bp:
        data = "mock0data"
        blob_stash = await bp.stash_as_pickle(data)
        assert await bp.blob_exists(blob_stash.filename)
        assert await bp.unstash_from_pickle(blob_stash.filename, blob_stash.md5sum) == data



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import contextvars
import logging
import os
import sys
import traceback
from collections import deque
from typing import TYPE_CHECKING, Any

from cachetools import TTLCache
from opentelemetry._logs import get_logger_provider
from opentelemetry.sdk.trace import ReadableSpan, SpanProcessor
from opentelemetry.trace import SpanContext, TraceFlags
from rich.logging import RichHandler
from rich.text import Text

from .auth import get_sync_token_credential

if TYPE_CHECKING:
    from azure.core.credentials import TokenCredential

TASK_LOG = logging.getLogger("task")
LOG = logging.getLogger("ai4s.jobq")
TRACK_LOG = logging.getLogger("ai4s.jobq.track")


def _azureml_run_description() -> dict[str, str]:
    """Returns a dict describing the AzureML run running the jobq worker."""
    if run_id := os.environ.get("AZUREML_RUN_ID"):
        workspace_scope = os.getenv("AZUREML_WORKSPACE_SCOPE")
        sub_id = os.environ.get("AZUREML_ARM_SUBSCRIPTION", "")
        resource_group = os.environ.get("AZUREML_ARM_RESOURCEGROUP", "")
        workspace_name = os.environ.get("AZUREML_ARM_WORKSPACE_NAME", "")
        return {
            "azureml_run_id": run_id,
            "azureml_workspace_name": workspace_name,
            "azureml_subscription_id": sub_id,
            "azureml_resource_group": resource_group,
            "azureml_project_name": os.environ.get("AZUREML_ARM_PROJECT_NAME", ""),
            "azureml_url": f"https://ml.azure.com/runs/{run_id}?wsid={workspace_scope}",
            "job_url": f"https://ml.azure.com/runs/{run_id}?wsid=/subscriptions/{sub_id}/resourcegroups/{resource_group}/workspaces/{workspace_name}",
        }
    # we are using the azureml_workspace_name in grafana dashboards to visualize queue status
    # which is why we set AZUREML_ARM_WORKSPACE_NAME also for e.g. azureml batch jobs
    return {
        "azureml_workspace_name": os.environ.get("AZUREML_ARM_WORKSPACE_NAME", ""),
        "job_url": os.environ.get("AZUREML_JOB_URL", ""),
    }


async def flush_app_insights():
    from ai4s.jobq.work import ProcessPoolRegistry

    await ProcessPoolRegistry.wait_for_msg_queue_to_drain()

    try:
        lp = get_logger_provider()
        if lp:
            if not getattr(lp, "force_flush", None):
                LOG.warning(
                    "Logger provider (%r) does not have force_flush() method: will not attempt to flush appinsights.",
                    lp.__class__.__name__,
                )
                return
            flushed = lp.force_flush()
            if not flushed:
                LOG.warning(
                    "force_flush() returned False: failed to flush application insights logs."
                )
            else:
                LOG.info("Flushed application insights logs.")
        else:
            LOG.warning("No logger provider found: cannot flush application insights logs.")
    except Exception as e:
        LOG.warning(f"Failed to flush application insights logs: {e}")


class SkipTaskLogsFilter(logging.Filter):
    """Filter out logs from task. modules."""

    def filter(self, record: logging.LogRecord) -> bool:
        return not record.name.startswith("task.")


_context_dimensions: contextvars.ContextVar[dict[str, str]] = contextvars.ContextVar(
    "_context_dimensions",
    default={},  # noqa: B039
)


class CustomDimensionsFilter(logging.Filter):
    """
    Adds custom 'extra' dimensions to log records (only if attribute is absent).
    Optionally drops records that carry exception context to prevent AppExceptions.
    """

    def __init__(self, custom_dimensions=None, *, filter_exceptions: bool = True):
        super().__init__()
        self.custom_dimensions = custom_dimensions or {}
        self.filter_exceptions = filter_exceptions

    def filter(self, record: logging.LogRecord) -> bool:
        if self.filter_exceptions:
            exc_info = getattr(record, "exc_info", None)
            if exc_info:
                exc_type, exc_value, exc_tb = exc_info

                # exceptions to traces to reduce AppExceptions table noise
                if exc_type:
                    tb_lines = traceback.format_exception(exc_type, exc_value, exc_tb)
                    tb_string = "".join(tb_lines)

                    record.exception_type = exc_type.__name__
                    record.exception_message = str(exc_value)
                    record.exception_traceback = tb_string

                    # clear exc_info to prevent it from going to AppExceptions table
                    record.exc_info = None
                    record.exc_text = None

        # Start with static (process-wide) dimensions
        cdim = self.custom_dimensions.copy()
        # Layer on per-coroutine context dimensions (override static ones)
        cdim.update(_context_dimensions.get())
        for key, value in cdim.items():
            if getattr(record, key, None) is not None:
                continue
            setattr(record, key, value)
        return True


_custom_dimensions_filter: CustomDimensionsFilter | None = None
_pending_custom_dimensions: dict[str, str] = {}


def set_custom_dimensions(**kwargs: str) -> None:
    """Add process-wide custom dimensions that will be included in all log records.

    These are shared across all async tasks in the current process.
    For per-coroutine dimensions (e.g. a unique worker_id per async worker),
    use ``set_context_dimensions()`` instead.

    Can be called before ``setup_logging()``; dimensions will be buffered and
    applied once logging is initialised.
    """
    if _custom_dimensions_filter is not None:
        _custom_dimensions_filter.custom_dimensions.update(kwargs)
    else:
        _pending_custom_dimensions.update(kwargs)


def set_context_dimensions(**kwargs: str) -> None:
    """Set per-coroutine custom dimensions that will be included in log records.

    Uses ``contextvars`` so each ``asyncio.Task`` gets its own copy.
    Values set here override the process-wide dimensions from ``set_custom_dimensions()``.
    """
    ctx = _context_dimensions.get().copy()
    ctx.update(kwargs)
    _context_dimensions.set(ctx)


class CachingLogHandler(logging.Handler):
    def __init__(self, **kwargs: Any):
        self._caches: TTLCache[str, deque] = TTLCache(maxsize=100, ttl=60)

        super().__init__(**kwargs)

    def get_log_cache(self, task_id: str) -> deque:
        logger_name = f"task.{task_id}"
        if logger_name not in self._caches:
            return deque()
        return self._caches[logger_name]

    def emit(self, record: logging.LogRecord) -> None:
        # not emitting anything, just caching
        if record.name.startswith("task."):
            if record.name not in self._caches:
                self._caches[record.name] = deque(maxlen=100)
            self._caches[record.name].append(record.msg)


class JobQRichHandler(RichHandler):
    def __init__(self, plain_task_logs: bool = False, **kwargs: Any):
        self.plain_task_logs = plain_task_logs

        super().__init__(**kwargs)

    def emit(self, record: logging.LogRecord) -> None:
        if self.plain_task_logs and record.name.startswith("task."):
            # no formatting (prefix with logger=task id, timestamp), just the raw job output here.
            print(record.getMessage(), flush=True)
        else:
            super().emit(record)

    def get_level_text(self, record: logging.LogRecord) -> Text:
        from rich.text import Text

        level_name = record.levelname
        level_text = level_name[:1] * 2
        styled_text = Text.styled(level_text, f"logging.level.{level_name.lower()}")
        return styled_text


class PlainHandler(logging.StreamHandler):
    def __init__(self, plain_task_logs: bool = False, **kwargs: Any):
        self.plain_task_logs = plain_task_logs
        kwargs.setdefault("stream", sys.stdout)
        super().__init__(**kwargs)

    def emit(self, record: logging.LogRecord) -> None:
        if self.plain_task_logs and record.name.startswith("task."):
            # no formatting (prefix with logger=task id, timestamp), just the raw job output here.
            print(record.msg, flush=True)
        else:
            super().emit(record)


class SkipHttpProcessor(SpanProcessor):
    """
    SpanProcessor that filters out HTTP spans to prevent them from being exported.
    """

    def on_end(self, span: ReadableSpan) -> None:
        """
        Called when a span ends. If the span's component is HTTP, it prevents the span from being exported.

        Args:
            span (ReadableSpan): The span that has ended.
        """
        if span._attributes is None:
            return

        if span._attributes.get("component") == "http":
            self._do_not_export(span)

    @staticmethod
    def _do_not_export(span: ReadableSpan) -> None:
        """
        Modifies the span context to ensure the span is not exported.

        This method sets the `TraceFlags` of the span's context to `DEFAULT`, which means
        that the span will not be sampled and thus not exported. The `TraceFlags` is a bitmask that
        includes a sampling bit to indicate whether the span should be sampled (exported) or not.

        Args:
            span (ReadableSpan): The span to modify.
        """
        span._context = SpanContext(
            span.context.trace_id,
            span.context.span_id,
            span.context.is_remote,
            TraceFlags(TraceFlags.DEFAULT),  # Set the TraceFlags to DEFAULT to prevent exporting
            span.context.trace_state,
        )


def setup_logging(
    queue_spec: str,
    app_insights_connection_string: str | None = None,
    environment: str | None = None,
    internal_log_level: int = logging.INFO,
    base_log_level: int = logging.WARNING,
) -> logging.Handler:
    """
    Sets up logging for jobq. This handles a few special cases, which a simple logging.basicConfig() call would not:

    * if application insights connection string is provided, it will set up Azure Monitor integration
      (this can also be done by setting the environment variable APPLICATIONINSIGHTS_CONNECTION_STRING)

    * if the environment variable JOBQ_ENVIRONMENT_NAME is set, it will add it to the custom dimensions, so logs can
      e.g. be filtered by the cluster a job is running on

    * depending on whether the input is a terminal or not, it will use a different log format.

    Args:

        queue_spec (str): The queue specification, ends up in the custom dimensions of the logs.
        base_log_level (int): The log level for non-ai4s.jobq logs (default: logging.WARNING).
        internal_log_level (int): The log level for internal logs (default: logging.INFO).
    """

    cache_log_handler = CachingLogHandler()
    # Include timestamps in log messages only when not in interactive terminal
    if not sys.stdin.isatty():
        log_handler: PlainHandler | JobQRichHandler = PlainHandler(plain_task_logs=True)
        log_handler.setLevel(internal_log_level)
        fmt = "%(asctime)s %(levelname)s: %(message)s [%(name)s]"
    else:
        log_handler = JobQRichHandler(plain_task_logs=True, show_path=False, show_time=False)
        fmt = "%(name)s: %(message)s"

    logging.basicConfig(
        level=base_log_level,
        format=fmt,
        datefmt="%Y-%m-%d %H:%M:%S",
        handlers=[cache_log_handler, log_handler],
    )

    connstr = app_insights_connection_string or os.environ.get(
        "APPLICATIONINSIGHTS_CONNECTION_STRING"
    )
    environment = environment or os.environ.get("JOBQ_ENVIRONMENT_NAME")

    global _custom_dimensions_filter
    custom_dims = {"queue": queue_spec} | _azureml_run_description()
    if environment:
        custom_dims["environment"] = environment
    custom_dims.update(_pending_custom_dimensions)
    _pending_custom_dimensions.clear()
    _custom_dimensions_filter = CustomDimensionsFilter(custom_dims)
    LOG.addFilter(_custom_dimensions_filter)
    TASK_LOG.addFilter(_custom_dimensions_filter)

    if connstr:
        try:
            from azure.monitor.opentelemetry import configure_azure_monitor

            try:
                credential: TokenCredential | None
                credential = get_sync_token_credential()
                # Use the Monitor scope for Application Insights with RBAC
                credential.get_token("https://monitor.azure.com/.default")
            except Exception as e:
                LOG.warning(
                    "Could not get a working token credential, setting up app insights without authentication"
                )
                LOG.debug(str(e))
                credential = None
        except ImportError:
            LOG.warning(
                "azure-monitor-opentelemetry cannot be imported. "
                "Please install it to enable Azure Monitor integration."
            )
        else:
            from opentelemetry.sdk._logs import (
                LoggingHandler as _SdkLoggingHandler,
            )

            try:
                from opentelemetry.instrumentation.logging.handler import (
                    LoggingHandler as _InstrLoggingHandler,
                )

                _logging_handler_types: tuple[type, ...] = (
                    _SdkLoggingHandler,
                    _InstrLoggingHandler,
                )
            except ImportError:
                _logging_handler_types = (_SdkLoggingHandler,)

            configure_azure_monitor(
                connection_string=connstr,
                span_processors=[SkipHttpProcessor()],
                credential=credential,
                sampling_ratio=0.0,
                enable_performance_counters=False,
                enable_live_metrics=False,
            )

            # Suppress noisy tracebacks from the OTel exporter when the
            # AppInsights endpoint is unreachable (e.g. after preemption).
            # The exporter logs ERROR + full traceback on every failed
            # export attempt, which can drown out real user output.
            logging.getLogger("azure.monitor.opentelemetry.exporter.export._base").setLevel(
                logging.CRITICAL
            )

            # Find the Azure Monitor LoggingHandler by type instead of assuming position
            azure_handler = None
            for handler in logging.getLogger().handlers:
                if isinstance(handler, _logging_handler_types):
                    azure_handler = handler
                    break

            if azure_handler is not None:
                azure_handler.addFilter(SkipTaskLogsFilter())
                azure_handler.addFilter(_custom_dimensions_filter)
            else:
                LOG.warning("Azure Monitor handler not found, custom filters not applied")

    LOG.setLevel(internal_log_level)
    TRACK_LOG.setLevel(internal_log_level)
    TASK_LOG.setLevel(internal_log_level)

    return log_handler
